#include <stdio.h>
#include <stdlib.h>


#define N 64
__uint128_t BINOMIAL[N+1][N+1];


/* Wikipedia or so ? :-) */
void generate_binomials() {
  int k, n;
  for (k = 1; k <= N; k++)
    BINOMIAL[0][k] = 0;
  for (n = 0; n <= N; n++)
    BINOMIAL[n][0] = 1;
  for (n = 1; n <= N; n++)
    for (k = 1; k <= N; k++)
      BINOMIAL[n][k] = BINOMIAL[n-1][k-1] + BINOMIAL[n-1][k];
}





/* Wikipedia Iterative version in C */
__uint128_t GCD(__uint128_t u, __uint128_t v) {
  int shift;

  /* GCD(0,v) == v; GCD(u,0) == u, GCD(0,0) == 0 */
  if (u == 0) return v;
  if (v == 0) return u;

  /* Let shift := lg K, where K is the greatest power of 2
        dividing both u and v. */
  for (shift = 0; ((u | v) & 1) == 0; ++shift) {
         u >>= 1;
         v >>= 1;
  }

  while ((u & 1) == 0)
    u >>= 1;

  /* From here on, u is always odd. */
  do {
    /* remove all factors of 2 in v -- they are not common */
    /*   note: v is not zero, so while will terminate */
    while ((v & 1) == 0)  /* Loop X */
      v >>= 1;

    /* Now u and v are both odd. Swap if necessary so u <= v,
       then set v = v - u (which is even). For bignums, the
       swapping is just pointer movement, and the subtraction
       can be done in-place. */
    if (u > v) {
      __uint128_t t = v; v = u; u = t;
    }  // Swap u and v.
    v = v - u;                       // Here v >= u.
  } while (v != 0);

  /* restore common factors of 2 */
  return u << shift;
}



__uint128_t  LCM(__uint128_t a, __uint128_t b) {
  return a/GCD(a,b) * b;
}


#define INT128_LONG_LONG(a) (long long unsigned)((a)>>64),(long long unsigned)((a)&~(((__uint128_t)1)<<64))

int main() {
  generate_binomials();
  __uint128_t lcm = 1;
  __uint128_t lcm_div_k_plus_1 = 0;
  long long int k_plus_1;
  printf ("#ifndef __BINOMIAL_WEIGHT__\n");
  printf ("#define __BINOMIAL_WEIGHT__\n");
  printf ("#define N_binomial_weight %d\n",N);
  printf ("__uint128_t binomial_weight[N_binomial_weight+1][N_binomial_weight+1] = {\n");
  for (k_plus_1 = 1; k_plus_1 <= N+1 ; k_plus_1++) {
    lcm = LCM(lcm,k_plus_1);

    /*
     * lcm { binomial(k,0) ; binomial(k,1) ; ... ; binomial(k,k) } = lcm {1;2;...;k;k+1} / (k+1)
     * "An identity involving the least common multiple of binomial coefficients and its application" Bakir FARHI
     */

    lcm_div_k_plus_1 = lcm / (k_plus_1);
    printf( "\t/* llcm{1..(k+1=%3lld)} = 0x%016llX%016llX \t llcm{1..(k+1=%3lld)}/(k+1=%3lld) = 0x%016llX%016llX */ {\n",
          k_plus_1,
          INT128_LONG_LONG(lcm),
          k_plus_1,k_plus_1,
          INT128_LONG_LONG(lcm_div_k_plus_1)
          );

    long long int i;
    long long int k = k_plus_1 - 1;
    for (i = 0; i <= k ; i++) {
      __uint128_t lcm_div_k_plus_1_div_binomial_k_i = lcm_div_k_plus_1/(BINOMIAL[k][i]);
      //__uint128_t lcm_div_k_plus_1_mod_binomial_k_i = lcm_div_k_plus_1%(BINOMIAL[k][i]);
      printf(
           //"\t /* llcm / Binomial{(k=%3lld),(i=%3lld)} = */ 0x%016llX%016llX /* [0x%016llX%016llX] */,\n"
           "\t\t /* llcm / Binomial{(k=%3lld),(i=%3lld)} = */ (((__uint128_t)0x%016llXULL)<<64) | 0x%016llXULL,\n"
           /* no support for 128bits literals in most compilers ... must be then build "by hand" from 64bits */
           , k, i
           , INT128_LONG_LONG(lcm_div_k_plus_1_div_binomial_k_i)
           //, INT128_LONG_LONG(lcm_div_k_plus_1_mod_binomial_k_i)
            );
    }
    printf( "\t},\n");
  }
  printf ("};\n");
  printf ("#endif\n");
  return 0;
}


